using namespace std; 

#include <vector>
#include <iostream>
#include <DataMat.h>
#include <NHLSOpt.h>
#include <EMDecomposer2D.h>
#include <string>
#include <fenv.h>
#include <fstream>
#include <cassert>
#include <sstream>
                        
using namespace std;
extern string OutputPath="./output" ;

inline double GetLum(double r, double g, double b) 
{ return (0.3*r+ 0.7*g + 0.01*b) ;}
 
inline void RGB2YUV(double r, double g, double b, double& Y, double& U, double& V) 
{
        double M[3][3] = {
                                {.299,          0.587,          0.114},
                                {-0.14713,      -0.28886,       0.436},
                                {0.615,         -0.51499,       -0.10001}
                         }; 
        Y=U=V = 0.0 ;
        double rgb[3] = {r,g,b} ;
        for (int i(0); i<3; i++)
        {
                Y += M[0][i] * rgb[i] ;
                U += M[1][i] * rgb[i] ;
                V += M[2][i] * rgb[i] ;
        }
        U=0.5*U/0.436 + 0.5 ;
        V=0.5*V/0.615 + 0.5 ;
}

inline void YUV2RGB(double y, double u, double v, double& R, double& G, double& B) 
{
        u= (u-0.5)*2*0.436;
        v= (v-0.5)*2*0.615;
        double M[3][3] = {
                {1,          0,          1.13983},
                {1,      -0.39465,       -0.58060},
                {1,         2.03211,       0}
        }; 
        
        R=G=B= 0.0 ;
        double yuv[3] = {y,u,v} ;
        for (int i(0); i<3; i++)
        {
                R += M[0][i] * yuv[i] ;
                G += M[1][i] * yuv[i] ;
                B += M[2][i] * yuv[i] ;
        }
}

 
void ReadImage(const string& image, DataMat& Rp, DataMat& Gp, DataMat& Bp, DataMat& InLum)
{        
        const double InvGamma = 2.2 ;
        const double Gamma = 1.0/InvGamma ;  
        const double LumEps = 0.01 ;                                
                
        string extn(image.substr(image.length()-3)); 
        // Save linearised R, G, B divided by Lum in Rp, Gp, Bp                         
        Magick::Image I;
        Magick::ColorRGB p;
        I.read(image);        
        if(I.size().width() == 0 || I.size().height() == 0 || !I.isValid())
        { cout << "Error loading " << image << endl; return; }
        
        int w = I.size().width();
        int h = I.size().height();
        int nR(h), nC(w) ;
        Rp.resize(nR, nC); Gp.resize(Rp); Bp.resize(Rp); InLum.resize(Rp);
                
        for(int x=0; x<w; x++)
        {                        
                for(int y=0; y<h; y++)
                {
                        p = I.pixelColor(x,y);
                        double r1(p.red()), g1(p.green()), b1(p.blue()) ;                                
                        double r(pow(r1, InvGamma)),g(pow(g1, InvGamma)), b(pow(b1, InvGamma)) ;
                        double Lum(GetLum(r,g,b)+LumEps) ;
                        InLum(y,x) = Lum ; Rp(y,x) = r/Lum ; Gp(y,x) = g/Lum ; Bp(y,x) = b/Lum ; 
                }
        }        
}   

void ImageDecomp(const string& image, int NModes, bool UseLogLum) 
{
        DataMat Rp, Gp, Bp, InLum ;        
        ReadImage(image, Rp, Gp, Bp, InLum) ; // Obtain R/L, G/L, B/L and L
        
        Rp.BinaryWrite("./output/BinData/RbyL.bin") ;
        Gp.BinaryWrite("./output/BinData/GbyL.bin") ;
        Bp.BinaryWrite("./output/BinData/BbyL.bin") ;
	
        int nR(InLum.NRows()), nC(InLum.NCols()) ;                        
        DataMat LogLum(nR, nC) ;
        const double logEps (0.1) ;
        	
        // Run Decomposition 
        EMDecomposer2D emd ;
        if (UseLogLum)
        {                
                for(int r(0); r<nR; r++) for(int c(0); c<nC; c++) LogLum(r,c) = log(InLum(r,c) + logEps) ;
                cout << " Decomposing in log domain " << endl ;
                emd.RunDecomposition(LogLum, NModes) ;
        }
        else 
                emd.RunDecomposition(InLum, NModes) ;
	
        const vector<DataMat>& modes= emd.GetIMFs() ;
        
        
        // Write output for each mode 
        for (int i(0); i<modes.size(); i++)
        {
                cout << "Writing Mode " << i << " of  ->" << modes.size() << endl ;  
                stringstream ss1 ; 
                ss1 << "./output/BinData/LumOut_" << i << ".bin" ; 
                modes[i].BinaryWrite(ss1.str()) ;                                
        }        
}

int main(int argc, char* argv[])
{
        feenableexcept(FE_DIVBYZERO) ;
        feenableexcept(FE_INVALID) ;
        
        string args ;
        for (int i(1) ; i<argc; i++) args = args + " " + argv[i] ;
        istringstream iss(args) ;
                
        string image ;                
        if (argc>1) iss >> image ;
        int NModes(1) ;        
        if (argc>2) iss >> NModes ;        
        string IsLog ;
        if (argc>3) iss >> IsLog ;                                
        bool UseLogLum (!(IsLog=="Linear" || IsLog=="linear")) ;
        ImageDecomp(image, NModes, UseLogLum) ;
        
	return -1;
}